from typing import Optional, Union

class Identifier:
    '''An auto-escaping identifier similar to a string.

    Conceptually immutable.
    '''
    __slots__ = (
        '__value',
        '__safe',
    )

    def __init__(self, value: Union['Identifier', str] = "") -> None:
        self.__safe: Optional[str]
        if isinstance(value, Identifier):
            self.__value = value.__value
            self.__safe = value.__safe
        else:
            self.__value = value
            self.__safe = None

    @property
    def value(self) -> str:
        return self.__value

    def __add__(self, other: Union['Identifier', str]) -> 'Identifier':
        if isinstance(other, Identifier):
            other = other.__value
        return Identifier(self.__value + other)

    def __radd__(self, other: Union['Identifier', str]) -> 'Identifier':
        if isinstance(other, Identifier):
            other = other.__value
        return Identifier(other + self.__value)

    def __contains__(self, other: Union['Identifier', str]) -> bool:
        if isinstance(other, Identifier):
            other = other.__value
        return self.__value.__contains__(other)

    def __hash__(self) -> int:
        return hash(self.__value)

    def __repr__(self) -> str:
        return f'<Identifier {self.__value!r}>'

    def __str__(self) -> str:
        if self.__safe is None:
            if b'\x00' in self.__value.encode('utf-8'):
                raise ValueError("sqlite Identifer must not contain any null bytes")

            self.__safe = '"' + self.__value.replace('"', '""') + '"'

        return self.__safe



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2021 Taylor C. Richberger
# This code is released under the license described in the LICENSE file

from contextlib import closing
import sqlite3
import unittest
from datetime import timedelta
from tempfile import TemporaryDirectory
from pathlib import Path
from tpcollections import Database, Mapping, MappingOrder
import json
import marshal
import pickle
import orjson

class TestExpiringDict(unittest.TestCase):
    def test_simple(self):
        with TemporaryDirectory() as temporary_directory:
            db_path = Path(temporary_directory) / 'test.db'

            with (
                Database(db_path) as db,
                db,
            ):
                d = Mapping(db)
                self.assertFalse(bool(d))
                self.assertEqual(tuple(d), ())
                self.assertEqual(tuple(d.keys()), ())
                self.assertEqual(tuple(d.items()), ())
                self.assertEqual(tuple(d.values()), ())
                self.assertEqual(len(d), 0)
                d['foo'] = 'bar'
                d['baz'] = 1337

            with (
                Database(db_path) as db,
                db,
            ):
                d = Mapping(db)
                self.assertTrue(bool(d))
                self.assertEqual(tuple(d), ('foo', 'baz'))
                self.assertEqual(tuple(d.keys()), ('foo', 'baz'))
                self.assertEqual(tuple(d.items()), (('foo', 'bar'), ('baz', 1337)))
                self.assertEqual(tuple(d.values()), ('bar', 1337))
                self.assertEqual(len(d), 2)

                self.assertEqual(tuple(reversed(d)), ('baz', 'foo'))
                self.assertEqual(tuple(reversed(d.keys())), ('baz', 'foo'))
                self.assertEqual(
                    tuple(reversed(d.items())),
                    (('baz', 1337), ('foo', 'bar')),
                )
                self.assertEqual(tuple(reversed(d.values())), (1337, 'bar'))

                self.assertEqual(tuple(d.keys(MappingOrder.KEY)), ('baz', 'foo'))
                self.assertEqual(
                    tuple(d.items(MappingOrder.KEY)),
                    (('baz', 1337), ('foo', 'bar')),
                )
                self.assertEqual(tuple(d.values(MappingOrder.KEY)), (1337, 'bar'))

                self.assertEqual(tuple(reversed(d.keys(MappingOrder.KEY))), ('foo', 'baz'))
                self.assertEqual(
                    tuple(reversed(d.items(MappingOrder.KEY))),
                    (('foo', 'bar'), ('baz', 1337)),
                )
                self.assertEqual(tuple(reversed(d.values(MappingOrder.KEY))), ('bar', 1337))

            with (
                Database(db_path) as db,
                db,
            ):
                d = Mapping(db)
                d['foo'] = 'barbar'

            with (
                Database(db_path) as db,
                db,
            ):
                d = Mapping(db)
                self.assertTrue(bool(d))
                self.assertEqual(tuple(d), ('foo', 'baz'))
                self.assertEqual(tuple(d.keys()), ('foo', 'baz'))
                self.assertEqual(tuple(d.items()), (('foo', 'barbar'), ('baz', 1337)))
                self.assertEqual(tuple(d.values()), ('barbar', 1337))
                self.assertEqual(len(d), 2)

            with (
                Database(db_path) as db,
                db,
            ):
                d = Mapping(db)
                del d['foo']

            with (
                Database(db_path) as db,
                db,
            ):
                d = Mapping(db)
                self.assertTrue(bool(d))
                self.assertEqual(tuple(d), ('baz',))
                self.assertEqual(tuple(d.keys()), ('baz',))
                self.assertEqual(tuple(d.items()), (('baz', 1337),))
                self.assertEqual(tuple(d.values()), (1337,))
                self.assertEqual(len(d), 1)

            with self.assertRaises(KeyError):
                with (
                    Database(db_path) as db,
                    db,
                ):
                    d = Mapping(db)
                    del d['foo']

            
            with (
                Database(db_path) as db,
                db,
            ):
                d = Mapping(db)
                d['foo'] = 'spam'

            with (
                Database(db_path) as db,
                db,
            ):
                d = Mapping(db)
                self.assertTrue(bool(d))
                self.assertEqual(tuple(d), ('baz', 'foo'))
                self.assertEqual(tuple(d.keys()), ('baz', 'foo'))
                self.assertEqual(tuple(d.items()), (('baz', 1337), ('foo', 'spam')))
                self.assertEqual(tuple(d.values()), (1337, 'spam'))
                self.assertEqual(len(d), 2)

    # def test_migration(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         with closing(sqlite3.connect(str(db_path))) as connection:
    #             with closing(connection.cursor()) as cursor:
    #                 cursor.execute('''
    #                     CREATE TABLE tpcollections (
    #                         key TEXT UNIQUE PRIMARY KEY NOT NULL,
    #                         expire INTEGER NOT NULL,
    #                         value BLOB NOT NULL
    #                     )
    #                 ''')
    #                 cursor.execute('''
    #                     CREATE INDEX tpcollections_expire_index
    #                         ON tpcollections (expire)
    #                 ''')
    #                 cursor.execute('''
    #                     CREATE TRIGGER tpcollections_insert_trigger
    #                         AFTER INSERT ON tpcollections
    #                     BEGIN
    #                         DELETE FROM tpcollections
    #                             WHERE expire <= strftime('%s', 'now');
    #                     END
    #                 ''')
    #                 cursor.execute('''
    #                     CREATE TRIGGER tpcollections_update_trigger
    #                         AFTER UPDATE ON tpcollections
    #                     BEGIN
    #                         DELETE FROM tpcollections
    #                             WHERE expire <= strftime('%s', 'now');
    #                     END
    #                 ''')
    #                 cursor.executemany(
    #                     '''
    #                         INSERT INTO tpcollections (key, expire, value)
    #                         VALUES (?, strftime('%s', 'now', '+7 days'), ?)
    #                     ''',
    #                     (
    #                         ('foo', '"bar"'),
    #                         ('baz', '1337'),
    #                     ),
    #                 )
    #             connection.commit()

    #         with SqliteDict(str(db_path)) as d:
    #             self.assertTrue(bool(d))
    #             self.assertEqual(set(d), {'foo', 'baz'})
    #             self.assertEqual(set(d.keys()), {'foo', 'baz'})
    #             self.assertEqual(set(d.items()), {('foo', 'bar'), ('baz', 1337)})
    #             self.assertEqual(set(d.values()), {'bar', 1337})
    #             self.assertEqual(len(d), 2)

    #         with SqliteDict(str(db_path)) as d:
    #             d['foo'] = 'barbar'

    #         with SqliteDict(str(db_path)) as d:
    #             self.assertTrue(bool(d))
    #             self.assertEqual(tuple(d), ('foo', 'baz'))
    #             self.assertEqual(tuple(d.keys()), ('foo', 'baz'))
    #             self.assertEqual(tuple(d.items()), (('foo', 'barbar'), ('baz', 1337)))
    #             self.assertEqual(tuple(d.values()), ('barbar', 1337))
    #             self.assertEqual(len(d), 2)

    #         with SqliteDict(str(db_path)) as d:
    #             del d['foo']

    #         with SqliteDict(str(db_path)) as d:
    #             self.assertTrue(bool(d))
    #             self.assertEqual(tuple(d), ('baz',))
    #             self.assertEqual(tuple(d.keys()), ('baz',))
    #             self.assertEqual(tuple(d.items()), (('baz', 1337),))
    #             self.assertEqual(tuple(d.values()), (1337,))
    #             self.assertEqual(len(d), 1)

    #         with self.assertRaises(KeyError):
    #             with SqliteDict(str(db_path)) as d:
    #                 del d['foo']

            
    #         with SqliteDict(str(db_path)) as d:
    #             d['foo'] = 'spam'

    #         with SqliteDict(str(db_path)) as d:
    #             self.assertTrue(bool(d))
    #             self.assertEqual(tuple(d), ('baz', 'foo'))
    #             self.assertEqual(tuple(d.keys()), ('baz', 'foo'))
    #             self.assertEqual(tuple(d.items()), (('baz', 1337), ('foo', 'spam')))
    #             self.assertEqual(tuple(d.values()), (1337, 'spam'))
    #             self.assertEqual(len(d), 2)

    # def test_quotes(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         with SqliteDict(
    #             str(db_path),
    #             table = 'Name\\with"special\t-_ char""ac"""ters',
    #         ) as d:
    #             self.assertFalse(bool(d))
    #             self.assertEqual(tuple(d), ())
    #             self.assertEqual(tuple(d.keys()), ())
    #             self.assertEqual(tuple(d.items()), ())
    #             self.assertEqual(tuple(d.values()), ())
    #             self.assertEqual(len(d), 0)
    #             d['foo'] = 'bar'
    #             d['baz'] = 1337

    #         with SqliteDict(
    #             str(db_path),
    #             table = 'Name\\with"special\t-_ char""ac"""ters',
    #         ) as d:
    #             self.assertTrue(bool(d))
    #             self.assertEqual(tuple(d), ('foo', 'baz'))
    #             self.assertEqual(tuple(d.keys()), ('foo', 'baz'))
    #             self.assertEqual(tuple(d.items()), (('foo', 'bar'), ('baz', 1337)))
    #             self.assertEqual(tuple(d.values()), ('bar', 1337))
    #             self.assertEqual(len(d), 2)

    # def test_autocommit(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         d = SimpleSqliteDict(str(db_path))
    #         self.assertFalse(bool(d))
    #         self.assertEqual(tuple(d), ())
    #         self.assertEqual(tuple(d.keys()), ())
    #         self.assertEqual(tuple(d.items()), ())
    #         self.assertEqual(tuple(d.values()), ())
    #         self.assertEqual(len(d), 0)
    #         d['foo'] = 'bar'
    #         d['baz'] = 1337

    #         d = SimpleSqliteDict(str(db_path))
    #         self.assertTrue(bool(d))
    #         self.assertEqual(tuple(d), ('foo', 'baz'))
    #         self.assertEqual(tuple(d.keys()), ('foo', 'baz'))
    #         self.assertEqual(tuple(d.items()), (('foo', 'bar'), ('baz', 1337)))
    #         self.assertEqual(tuple(d.values()), ('bar', 1337))
    #         self.assertEqual(len(d), 2)

    # def test_isolation_level(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         d = SimpleSqliteDict(str(db_path), isolation_level='DEFERRED')
    #         self.assertFalse(bool(d))
    #         self.assertEqual(tuple(d), ())
    #         self.assertEqual(tuple(d.keys()), ())
    #         self.assertEqual(tuple(d.items()), ())
    #         self.assertEqual(tuple(d.values()), ())
    #         self.assertEqual(len(d), 0)
    #         d['foo'] = 'bar'
    #         d.connection.commit()
    #         d['baz'] = 1337
    #         d.connection.rollback()

    #         d = SimpleSqliteDict(str(db_path))
    #         self.assertTrue(bool(d))
    #         self.assertEqual(tuple(d), ('foo',))
    #         self.assertEqual(tuple(d.keys()), ('foo',))
    #         self.assertEqual(tuple(d.items()), (('foo', 'bar'),))
    #         self.assertEqual(tuple(d.values()), ('bar',))
    #         self.assertEqual(len(d), 1)

    # def test_nested(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         with SqliteDict(str(db_path)) as d:
    #             d['foo'] = {'foo': 'bar', 'baz': [2, 'two']}

    #         with SqliteDict(str(db_path)) as d:
    #             self.assertEqual(d['foo'], {'foo': 'bar', 'baz': [2, 'two']})

    # def test_json(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         with SqliteDict(str(db_path), serializer = json) as d:
    #             d['foo'] = {'foo': 'bar', 'baz': [2, 'two']}

    #         with SqliteDict(str(db_path), serializer = json) as d:
    #             self.assertEqual(d['foo'], {'foo': 'bar', 'baz': [2, 'two']})

    # def test_orjson(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         with SqliteDict(str(db_path), serializer = orjson) as d:
    #             d['foo'] = {'foo': 'bar', 'baz': [2, 'two']}

    #         with SqliteDict(str(db_path), serializer = orjson) as d:
    #             self.assertEqual(d['foo'], {'foo': 'bar', 'baz': [2, 'two']})

    # def test_pickle(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         with SqliteDict(str(db_path), serializer = pickle) as d:
    #             d['foo'] = {'foo': 'bar', 'baz': [2, 'two']}

    #         with SqliteDict(str(db_path), serializer = pickle) as d:
    #             self.assertEqual(d['foo'], {'foo': 'bar', 'baz': [2, 'two']})

    # def test_marshal(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         with SqliteDict(str(db_path), serializer = marshal) as d:
    #             d['foo'] = {'foo': 'bar', 'baz': [2, 'two']}

    #         with SqliteDict(str(db_path), serializer = marshal) as d:
    #             self.assertEqual(d['foo'], {'foo': 'bar', 'baz': [2, 'two']})

    # def test_expire(self):
    #     with TemporaryDirectory() as temporary_directory:
    #         db_path = Path(temporary_directory) / 'test.db'

    #         with SqliteDict(str(db_path)) as d:
    #             d['foo'] = 'bar'
    #             d['postponed'] = 'worked'

    #         with SqliteDict(str(db_path)) as d:
    #             d.lifespan = timedelta(weeks=-1)
    #             d.postpone_all()
    #             d.lifespan = timedelta(weeks=1)
    #             d.postpone('postponed')
    #             # This triggers the actual expiry
    #             d['baz'] = 1337

    #         with SqliteDict(str(db_path)) as d:
    #             self.assertTrue(bool(d))
    #             self.assertEqual(tuple(d), ('postponed', 'baz'))
    #             self.assertEqual(tuple(d.keys()), ('postponed', 'baz'))
    #             self.assertEqual(
    #                 tuple(d.items()),
    #                 (('postponed', 'worked'), ('baz', 1337)),
    #             )
    #             self.assertEqual(tuple(d.values()), ('worked', 1337))
    #             self.assertEqual(len(d), 2)

if __name__ == '__main__':
    unittest.main()



import json as _json
from typing import Callable, NamedTuple, Union, Any
from functools import partial
import pickle as _pickle

class Serializer(NamedTuple):
    loads: Callable[[Any], Any]
    dumps: Callable[[Any], Any]

json = Serializer(
    loads=_json.loads,
    dumps=partial(_json.dumps, separators=(',', ':')),
)

pickle = Serializer(
    loads=_pickle.loads,
    dumps=partial(_pickle.dumps, protocol=_pickle.HIGHEST_PROTOCOL),
)

deterministic_json = Serializer(
    loads=_json.loads,
    dumps=partial(_json.dumps, sort_keys=True, separators=(',', ':')),
)

try:
    import orjson as _orjson

    orjson = Serializer(
        loads=_orjson.loads,
        dumps=lambda value: _orjson.dumps(value).decode('utf-8'),
    )
    deterministic_orjson = Serializer(
        loads=_orjson.loads,
        dumps=lambda value: _orjson.dumps(value, option=_orjson.OPT_SORT_KEYS).decode('utf-8'),
    )
except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This code is distributed under the terms and conditions
# from the Apache License, Version 2.0
#
# http://opensource.org/licenses/apache2.0.php
#
# This code was inspired by:
#  * http://code.activestate.com/recipes/576638-draft-for-an-sqlite3-based-dbm/
#  * http://code.activestate.com/recipes/526618/


import json
import sqlite3
from collections.abc import MutableMapping
from contextlib import ExitStack, closing, contextmanager
from datetime import timedelta
from types import TracebackType
from typing import Any, Generator, Iterable, Iterator, Optional, Reversible, Tuple, Type, Union
from weakref import finalize
from enum import unique, Enum

from ._util import Identifier

from . import _db, _serializers

@unique
class MappingOrder(str, Enum):
    '''An ordering enum for iteration methods.
    '''

    ID = 'id'
    KEY = 'key'

    def __str__(self) -> str:
        return self.value

    def __format__(self, format_spec: str) -> str:
        return self.value.__format__(format_spec)

class _Keys(Reversible, Iterable[Any]):
    __slots__ = (
        '_connection',
        '_database',
        '_table',
        '_serializer',
        '_order',
    )

    def __init__(
        self,
        connection: _db.Connection,
        database: Identifier,
        table: Identifier,
        serializer: _serializers.Serializer,
        order: MappingOrder,
    ) -> None:

        self._connection = connection
        self._database = database
        self._table = table
        self._serializer = serializer
        self._order = order
    
    def _iterator(self, order: str) -> Iterator[Any]:
        with self._connection.cursor() as cursor:
            for key, in cursor.execute(
                f'SELECT key FROM {self._database}.{self._table} ORDER BY {self._order} {order}',
            ):
                yield self._serializer.loads(key)

    def __iter__(self) -> Iterator[Any]:
        return self._iterator('ASC')

    def __reversed__(self) -> Iterator[Any]:
        return self._iterator('DESC')

class _Values(Reversible, Iterable[Any]):
    __slots__ = (
        '_connection',
        '_database',
        '_table',
        '_serializer',
        '_order',
    )

    def __init__(
        self,
        connection: _db.Connection,
        database: Identifier,
        table: Identifier,
        serializer: _serializers.Serializer,
        order: MappingOrder,
    ) -> None:

        self._connection = connection
        self._database = database
        self._table = table
        self._serializer = serializer
        self._order = order

    def _iterator(self, order: str) -> Iterator[Any]:
        with self._connection.cursor() as cursor:
            for value, in cursor.execute(
                f'SELECT value FROM {self._database}.{self._table} ORDER BY {self._order} {order}',
            ):
                yield self._serializer.loads(value)

    def __iter__(self) -> Iterator[Any]:
        return self._iterator('ASC')

    def __reversed__(self) -> Iterator[Any]:
        return self._iterator('DESC')

class _Items(Reversible, Iterable[Tuple[Any, Any]]):
    __slots__ = (
        '_connection',
        '_database',
        '_table',
        '_key_serializer',
        '_value_serializer',
        '_order',
    )

    def __init__(
        self,
        connection: _db.Connection,
        database: Identifier,
        table: Identifier,
        key_serializer: _serializers.Serializer,
        value_serializer: _serializers.Serializer,
        order: MappingOrder,
    ) -> None:
        self._connection = connection
        self._database = database
        self._table = table
        self._key_serializer = key_serializer
        self._value_serializer = value_serializer
        self._order = order
    
    def _iterator(self, order: str) -> Iterator[Tuple[Any, Any]]:
        with self._connection.cursor() as cursor:
            for key, value in cursor.execute(f'''
                SELECT key, value FROM {self._database}.{self._table}
                    ORDER BY {self._order} {order}
            '''):
                yield (
                    self._key_serializer.loads(key),
                    self._value_serializer.loads(value),
                )

    def __iter__(self) -> Iterator[Tuple[Any, Any]]:
        return self._iterator('ASC')

    def __reversed__(self) -> Iterator[Tuple[Any, Any]]:
        return self._iterator('DESC')

class Mapping(_db._Base, MutableMapping):
    '''A database mapping.
    '''

    __slots__ = (
        '_key_serializer',
        '_value_serializer',
    )

    def __init__(self,
        connection: _db.Connection,
        database: str = 'main',
        table: str = 'mapping',
        key_serializer: _serializers.Serializer = _serializers.deterministic_json,
        value_serializer: _serializers.Serializer = _serializers.pickle,
    ) -> None:

        super().__init__(connection, Identifier(database), Identifier(table), 'mapping')

        self._key_serializer = key_serializer
        self._value_serializer = value_serializer

        version = self._version
        previous_version = version

        if version < 1:
            with self._connection.cursor() as cursor:
                cursor.execute(f'''
                    CREATE TABLE {self._database}.{self._table} (
                        id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
                        key {_db.ANY} UNIQUE NOT NULL,
                        value {_db.ANY} NOT NULL) {_db.STRICT}
                ''')
                version = 1

        if version > 1:
            raise ValueError('tpcollections is not forward compatible')

        if version != previous_version:
            self._version = version

    def __bool__(self) -> bool:
        '''Check if the table is not empty.'''

        return len(self) > 0

    def keys(self, order: MappingOrder = MappingOrder.ID) -> Reversible[Any]:
        '''Iterate over keys in the table.
        '''

        return _Keys(
            connection=self._connection,
            database=self._database,
            table=self._table,
            serializer=self._key_serializer,
            order=order,
        )

    def __iter__(self) -> Iterator[Any]:
        return iter(self.keys())

    def __reversed__(self) -> Iterator[Any]:
        return reversed(self.keys())

    def values(self, order: MappingOrder = MappingOrder.ID) -> Reversible[Any]:
        '''Iterate over values in the table.
        '''

        return _Values(
            connection=self._connection,
            database=self._database,
            table=self._table,
            serializer=self._value_serializer,
            order=order,
        )

    def items(self, order: MappingOrder = MappingOrder.ID) -> Reversible[Tuple[Any, Any]]:
        '''Iterate over keys and values in the table.
        '''

        return _Items(
            connection=self._connection,
            database=self._database,
            table=self._table,
            key_serializer=self._key_serializer,
            value_serializer=self._value_serializer,
            order=order,
        )

    def __contains__(self, key: Any) -> bool:
        '''Check if the table contains the given key.
        '''

        with self._connection.cursor() as cursor:
            cursor.execute(
                f'SELECT 1 FROM {self._database}.{self._table} WHERE key = ?',
                (self._key_serializer.dumps(key),),
            )
            return cursor.fetchone() is not None

    def __getitem__(self, key: str) -> Any:
        '''Fetch the key.
        '''

        with self._connection.cursor() as cursor:
            for row in cursor.execute(
                f'SELECT value FROM {self._database}.{self._table} WHERE key = ?',
                (self._key_serializer.dumps(key),),
            ):
                return self._value_serializer.loads(row[0])
        raise KeyError(key)

    def __setitem__(self, key: str, value: Any) -> None:
        '''Set or replace the item.

        This also triggers cleaning up expired values.
        '''

        with self._connection.cursor() as cursor:
            if sqlite3.sqlite_version_info >= (3, 24):
                cursor.execute(f'''
                        INSERT INTO {self._database}.{self._table} (key, value)
                            VALUES (?, ?)
                            ON CONFLICT (key) DO UPDATE
                            SET value=excluded.valu
                    ''',
                    (
                        self._key_serializer.dumps(key),
                        self._value_serializer.dumps(value),
                    ),
                )
            elif key in self:
                cursor.execute(f'''
                        UPDATE {self._database}.{self._table}
                            SET value=?2
                            WHERE key=?1
                    ''',
                    (
                        self._key_serializer.dumps(key),
                        self._value_serializer.dumps(value),
                    ),
                )
            else:
                cursor.execute(f'''
                        INSERT INTO {self._database}.{self._table} (key, value)
                            VALUES (?, ?)
                    ''',
                    (
                        self._key_serializer.dumps(key),
                        self._value_serializer.dumps(value),
                    ),
                )

    def __delitem__(self, key: str) -> None:
        '''Delete an item from the table.
        '''

        with self._connection.cursor() as cursor:
            cursor.execute(
                f'DELETE FROM {self._database}.{self._table} WHERE key=?',
                (self._key_serializer.dumps(key),),
            )
            if cursor.rowcount != 1:
                raise KeyError(key)

    def clear(self) -> None:
        '''Delete all items from the table.
        '''

        with self._connection.cursor() as cursor:
            cursor.execute(f'DELETE FROM {self._database}.{self._table}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import sqlite3
from collections.abc import MutableMapping
from functools import wraps
from contextlib import ExitStack, closing, contextmanager
from datetime import timedelta
from pathlib import Path
from types import TracebackType
from typing import Any, ContextManager, Generator, Iterable, Iterator, List, Optional, Reversible, Set, Tuple, Type, Union
from weakref import finalize
from enum import auto, unique, Enum
import warnings

from ._util import Identifier

@contextmanager
def _savepoint(
    connection: sqlite3.Connection,
    name: Identifier | str = Identifier('tpcollection'),
) -> Generator[None, None, None]:
    if isinstance(name, str):
        name = Identifier(name)

    with closing(connection.cursor()) as cursor:
        cursor.execute(f'SAVEPOINT {name}')
        try:
            yield
        except:
            cursor.execute(f'ROLLBACK TO {name}')
            raise
        finally:
            cursor.execute(f'RELEASE {name}')

@contextmanager
def _transaction(
    connection: sqlite3.Connection,
    read_only: bool = False,
) -> Generator[None, None, None]:
    with closing(connection.cursor()) as cursor:
        if read_only:
            cursor.execute('BEGIN')
        else:
            cursor.execute('BEGIN IMMEDIATE')

        try:
            yield
        except:
            cursor.execute('ROLLBACK')
            raise
        else:
            cursor.execute('COMMIT')

@unique
class Mode(Enum):
    # Database may be read or written by this connection.
    READ_WRITE = auto()

    # Database may be read by this connection but not written.  Other
    # connections may write to it, and this connection will reflect those
    # changes.  This connection may still establish locks, and therefore
    # might need to write to the filesystem. This mode is mostly useful
    # for performance reasons, as various read-only connections may read in
    # parallel.
    READ_ONLY = auto()

    # The database will not have any writes or locks done to it or its
    # filesystem.  Other connections are expected to not write to the database
    # at all.
    IMMUTABLE = auto()

def _uri(
    path: Optional[Path] = None,
    mode: Mode = Mode.READ_WRITE,
) -> str:
    if path is None:
        return ':memory:'
    else:
        if path.is_absolute():
            uri = path.as_uri()
        else:
            uri = 'file:' + str(path)

        if mode is Mode.READ_ONLY:
            uri += '?mode=ro'
        elif mode is Mode.IMMUTABLE:
            uri += '?immutable=1'

        return uri

class Database:
    """
    A sqlite connection manager.

    This can be called to return a simple context manager that opens and closes
    a connection (useful for getting connection per thread), or it can be used
    directly as a context manager.

    If any nestable or concurrent use is desired, this must be called.  It is
    not otherwise re-entrant.
    """

    __slots__ = (
        '_uri',
        '_mode',
        '_connection',
        '_timeout',
        '_wal',
        '__weakref__'
    )

    def __init__(self,
        path: Optional[Path] = None,
        mode: Mode = Mode.READ_WRITE,
        wal: bool = False,
        timeout: float = 5.0,
        **kwargs,
    ) -> None:
        assert not (path is None and mode is not Mode.READ_WRITE), (
            'An in-memory database must be read-write'
        )

        self._mode = mode
        self._uri = _uri(path, mode)
        self._wal = wal
        self._timeout = timeout

    @property
    def read_only(self) -> bool:
        return self._mode is not Mode.READ_WRITE

    def __call__(self) -> ContextManager['Connection']:
        return _connect(
            self._uri,
            self._mode,
            self._wal,
            self._timeout,
        )

    def __enter__(self) -> 'Connection':
        assert not hasattr(self, '_connection'), (
            'Database is not a nestable context manager, call it instead'
        )
        self._connection = self()
        return self._connection.__enter__()

    def __exit__(
        self,
        type: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        try:
            return self._connection.__exit__(type, value, traceback)
        finally:
            del self._connection


STRICT: str = ''
ANY: str = 'BLOB'
WITHOUT_ROWID: str = ''

if sqlite3.sqlite_version_info >= (3, 37):
    STRICT = 'STRICT'
    ANY = 'ANY'

if sqlite3.sqlite_version_info >= (3, 8, 2):
    WITHOUT_ROWID = 'WITHOUT ROWID'

if sqlite3.sqlite_version_info >= (3, 38):
    UNIXEPOCH = 'UNIXEPOCH()'
else:
    UNIXEPOCH = "CAST(strftime('%s', 'now') AS INTEGER)"

_APPLICATION_ID = -1238962565

_STRICT_WITHOUT_ROWID = ', '.join(part for part in (STRICT, WITHOUT_ROWID) if part)

class Connection(MutableMapping):
    '''The actual connection object, as a MutableMapping[str, Any].

    Items are expired when a value is inserted or updated.  Deletion or
    postponement does not expire items.
    '''

    __slots__ = (
        '_connection',
        '_attachments',
        '_mode',
        '_transactions',
        '_wal',
        '__weakref__',
    )

    def __init__(self,
        connection: sqlite3.Connection,
        mode: Mode,
        wal: bool,
    ) -> None:
        self._connection = connection
        self._attachments: Set[str] = {'main'}
        self._mode = mode
        self._transactions: List[ContextManager[None]] = []
        self._wal = wal
        self._init(Identifier('main'))

    def _init(self, database: Identifier) -> None:
        with self.cursor() as cursor:
            application_id = next(cursor.execute('PRAGMA application_id'))[0]
            if application_id == 0:
                cursor.execute(f'PRAGMA application_id = {_APPLICATION_ID}')
            elif application_id != _APPLICATION_ID:
                raise ValueError(f'illegal application ID {application_id}')

            user_version = next(cursor.execute('PRAGMA user_version'))[0]
            if user_version == 0:
                cursor.execute(f'PRAGMA user_version = 1')
            elif user_version != 1:
                raise ValueError(f'user_version was {user_version}')

            if not self.read_only:
                cursor.execute(f'''
                    CREATE TABLE IF NOT EXISTS main.tpcollections (
                        name TEXT PRIMARY KEY NOT NULL,
                        type TEXT NOT NULL,
                        version INTEGER NOT NULL
                    ) {_STRICT_WITHOUT_ROWID}
                ''')

    @property
    def connection(self) -> sqlite3.Connection:
        return self._connection

    @contextmanager
    def cursor(self) -> Generator[sqlite3.Cursor, None, None]:
        with closing(self._connection.cursor()) as cursor:
            yield cursor

    @property
    def read_only(self) -> bool:
        return self._mode is not Mode.READ_WRITE

    def __enter__(self) -> None:
        if self._transactions:
            new_transaction = _savepoint(self._connection)
        else:
            new_transaction = _transaction(self._connection, self.read_only)

        self._transactions.append(new_transaction)

    def __exit__(
        self,
        type: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        return self._transactions.pop().__exit__(type, value, traceback)

    def attach(
        self,
        database: str,
        path: Optional[Path] = None,
    ) -> None:
        uri = _uri(path, self._mode)
        database_id = Identifier(database)

        with closing(self._connection.cursor()) as cursor:
            if __debug__ and self._transactions:
                warnings.warn(
                    'Attaching a database inside a transaction can prevent '
                    'transactions from being atomic.'
                )
            cursor.execute(f'ATTACH ? AS {database_id}', (uri,))
            if self._wal and not self.read_only:
                if __debug__:
                    warnings.warn(
                        'Attaching a database in wal mode can cause'
                        ' non-atomic transactions:'
                        ' https://www.sqlite.org/lang_attach.html'
                    )
                cursor.execute(f'PRAGMA {database_id}.journal_mode=WAL')
                cursor.execute(f'PRAGMA {database_id}.synchronous=NORMAL')

        self._init(database_id)

@contextmanager
def _connect(
    uri: str,
    mode: Mode,
    wal: bool,
    timeout: float,
) -> Generator[Connection, None, None]:
    with (
        closing(sqlite3.connect(uri, timeout=timeout)) as connection,
        closing(connection.cursor()) as cursor,
    ):
        try:
            if wal and mode is Mode.READ_WRITE:
                cursor.execute('PRAGMA main.journal_mode=WAL')
                cursor.execute('PRAGMA main.synchronous=NORMAL')

            yield Connection(connection, mode, wal)
        finally:
            if mode is Mode.READ_WRITE:
                cursor.execute('PRAGMA analysis_limit=8192')
                cursor.execute('PRAGMA optimize')

class _Base:
    __slots__ = (
        '_connection',
        '_database',
        '_table',
    )

    def __init__(
        self,
        connection: Connection,
        database: Identifier,
        table: Identifier,
        type: str,
    ) -> None:
        self._connection = connection
        self._database = database
        self._table = table

        with closing(connection.connection.cursor()) as cursor:
            cursor.execute(f'''
                SELECT type
                    FROM {database}.tpcollections
                    WHERE name = ?
            ''', (table.value,))
            row = cursor.fetchone()
            if row is None:
                cursor.execute(
                    'SELECT 1 FROM sqlite_master WHERE name = ?',
                    (table.value,)
                )
                if cursor.fetchone() is not None:
                    raise NameError(f'table {table} already exists')
                
                cursor.execute(f'''
                    INSERT INTO {database}.tpcollections 
                        (name, type, version)
                        VALUES (?, ?, ?)
                ''', (table.value, type, 0))
            else:
                existing_type, = row

                if type != existing_type:
                    raise ValueError(f'Tried to open {database}.{table}'
                        f' as {type}, but it already existed as {existing_type}')
    @property
    def database(self) -> str:
        return self._database.value

    @property
    def table(self) -> str:
        return self._table.value

    @property
    def _version(self) -> int:
        with closing(self._connection.connection.cursor()) as cursor:
            cursor.execute(f'''
                SELECT version
                    FROM {self._database}.tpcollections
                    WHERE name = ?
            ''', (self._table.value,))
            version, = cursor.fetchone()
            return version

    @_version.setter
    def _version(self, value: int) -> None:
        assert not self._connection.read_only
        with closing(self._connection.connection.cursor()) as cursor:
            cursor.execute(f'''
                UPDATE {self._database}.tpcollections
                    SET version = ?
                    WHERE name = ?
            ''', (value, self._table.value))

